class ScriptRunner

  def initialize(options)
    fail 'You need to load this script through a require' unless __FILE__
    
    @options = options
    @options[:read_only] = true unless @options.key?(:read_only)
    @options[:skip] = 0  unless @options.key?(:skip)
  end 

  def run
    @start_time = DateTime.current
    puts "Start script at #{@start_time}"
    process_rows

    @end_time = DateTime.current
    puts "End script at #{@end_time}"
    puts "Executed in #{print_duration(@start_time, @end_time)}"
    puts "Log filename : #{@log_file_name}"

    generate_stats
  end

  private

  def print_duration(start_time, end_time)
    difference = end_time.to_f - start_time.to_f

    days = (difference / (24 * 60 * 60)).to_i
    difference %= (24 * 60 * 60)

    hours = (difference / (60 * 60)).to_i
    difference %= (60 * 60)

    minutes = (difference / 60).to_i
    seconds = (difference % 60).to_i

    str = "#{seconds}s"
    str = "#{minutes}m #{str}" if minutes > 0
    str = "#{hours}h #{str}" if hours > 0
    str = "#{days}d #{str}" if days > 0
    str
  end 

  def add_log_line(log_line, return_code, modified_fields, code_details)
    full_line = log_line.fill(nil, log_line.length...HEADERS.length) + [return_code, modified_fields&.join(', '), code_details]
    @log_file << full_line
    puts full_line.join('; ')
  end

  def add_stats_line(stats_line)
    @stats_file << stats_line
    puts stats_line.join('; ')
  end 

  def generate_filename(type, page_nb = nil)
    page_string = page_nb ? "_page_#{page_nb}" : ''
    "#{HERMES_CONFIG['environment']}-#{@start_time.strftime('%Y-%m-%dT%H:%M:%S')}_#{File.basename(__FILE__, '.rb')}_#{type}#{@options[:read_only] ? '_ro' : ''}#{page_string}.csv"
  end

  def print_percent(value, total)
    "#{(value.to_f / total * 100).round(2)}%"
  end

  def generate_stats
    processed_element_count = ok_nb = ko_nb = nok_nb = 0
    start_time = DateTime.current
    puts "Start generating stats at #{start_time}"
    errors = {}
    ids = {}

    general_file_name = generate_filename('logs').sub('.csv', '')
    Dir.glob("#{general_file_name}*.csv").each do |file_path|
      CSV.foreach(file_path, 'rb', col_sep: ";").with_index do |line, index|
        next if index == 0 
        processed_element_count += 1
        case line[HEADERS.length]
        when 'OK'
          ok_nb += 1
        when 'KO'
          ko_nb += 1
        when 'NOK'
          nok_nb += 1
        end

        if error_code = line[HEADERS.length + 2] 
          errors[error_code] ||= 0
          errors[error_code] += 1
        end 

        ids[line.first] ||= 0
        ids[line.first] += 1
      end
    end

    stats_file_name = generate_filename('stats')
    dublicated_id_nb = ids.select { |k, v| v > 1 }.length

    CSV.open("./#{stats_file_name}", 'ab', col_sep: ";") do |stats_file|
      @stats_file = stats_file

      add_stats_line(['Statistiques :'])

      add_stats_line([
        'Elements inspectes',
        @inspected_element_count,
      ])

      add_stats_line([
        'Elements traites',
        processed_element_count,
        "100%",
      ])

      add_stats_line([
        'OK',
        ok_nb,
        print_percent(ok_nb, processed_element_count)
      ])

      add_stats_line([
        'KO',
        ko_nb,
        print_percent(ko_nb, processed_element_count)
      ])

      add_stats_line([
        'NOK',
        nok_nb,
        print_percent(nok_nb, processed_element_count)
      ])

      add_stats_line([
        'Nombre d\'id dupliques',
        dublicated_id_nb,
        print_percent(dublicated_id_nb, processed_element_count)
      ])
      
      add_stats_line([])
      add_stats_line(['Erreurs :'])

      if errors.any?
        errors.each do |key, value|
          add_stats_line([
            key,
            value,
            "#{(value.to_f / processed_element_count * 100).round(2)}%",
          ])
        end
      else
        add_stats_line([
          'Aucune erreur',
        ])
      end

      add_stats_line([])
      add_stats_line(['Execution du script :'])
      add_stats_line([
        'Date de debut',
        @start_time
      ])
      add_stats_line([
        'Date de fin',
        @end_time
      ])
      add_stats_line([
        'Duree',
        print_duration(@start_time, @end_time)
      ])

      # You can add stats here
      # add_stats_line([])
      # add_stats_line(['Autres statistique :'])
    end

    end_time = DateTime.current
    puts "End generating stats at #{end_time}"
    puts "Stats generated in  #{print_duration(start_time, end_time)}"
    puts "Stats filename : #{stats_file_name}"
  end 

  HEADERS = [
    "housing_request_id",
    "als_id",
    "unique_number",
    "updated_at",
    "active_application_id",
    "main_applicant_id",
    'main_applicant + coapplicants dynamic_amount_minus_two',
    'main_applicant + coapplicants dynamic_amount_minus_one',
    'main_applicant + coapplicants amount_year_minus_two',
    'main_applicant + coapplicants amount_year_minus_one',
    'old_reference_tax_revenues',
    'reference_tax_revenues',
    'old_request_status',
    'request_status',
    'old_required_document_status',
    'required_document_status',
  ]

  def process_rows
    per_page = 50_000
    page = 1
    continue_paginate = true

    ids = Hermes::HousingRequest.where(is_active: true, :request_status.nin => ['ANNULEE', 'SATISFAITE']).order_by(id: :asc).pluck(:id)

    @inspected_element_count = ids.length

    while continue_paginate  do
      @log_file_name = generate_filename('logs', page)
      start_index = (page - 1) * per_page + @options[:skip]
      page_ids = ids[start_index, per_page]

      CSV.open("./#{@log_file_name}", 'ab', col_sep: ";") do |log_file|
        @log_file = log_file
        @log_file << HEADERS + ['return_code', 'modified_fields', 'code_details'] 
      
        r = Hermes::HousingRequest.where(:id.in=> page_ids).batch_size(1000).order_by(id: :asc)
      
        r.no_timeout.each_with_index do |record, i|
          puts "#{i + start_index}/#{@inspected_element_count}" if i % 1000 == 0
          process_single_row(record)
        end
      end

      continue_paginate = @inspected_element_count >= start_index + per_page
      page += 1
    end 
  end

  def get_required_document_status(hrf)
    filtered_hashes = hrf.required_documents.select { |hash| hash[:type].include?("imposition") }
    filtered_hashes.map { |hash| "#{hash[:applicant_id]} (#{hash[:type][-4..-1]})" }
  end 

  def validate_fields(hr)
    path = hr.form_validator_csv_path
    rule_to_checks = CSV.read(path)
    version = rule_to_checks[0][1]
    front_errors = []
    sne_errors = []
    rule_to_checks.each_with_index do |line_to_check,i|
        if(i>1)
            unless hr.check_rule_to_check_line_to_check(line_to_check)
                if line_to_check[6] == "1"
                    sne_errors.push(hr.format_error(line_to_check, i, version))
                else
                    front_errors.push(hr.format_error(line_to_check, i, version))
                end
            end
        end
    end
   
    hr.form_front_validation_errors = front_errors
    hr.form_sne_validation_errors = sne_errors
  end

  def process_single_row(record)
    log_line = modified_fields = code_details = return_code = nil
    return_code = 'OK'

    hr = record
    hrf = hr.housing_request_folder
    return unless hrf

    all_applicant_ids = [hr.main_applicant_id.to_s] + (hr.get_applicant_ids.map(&:to_s) || []) 
    old_reference_tax_revenues = hr.reference_tax_revenues
    old_request_status = hr.request_status 
    old_required_document_status = get_required_document_status(hrf)
    
    hrf.required_documents = hrf.required_documents.reject { |hash| hash[:type].include?("imposition") }

    all_applicant_ids.each do |applicant_id|
      hrf.calculate_revenus_fiscaux(applicant_id)
    end

    unless @options[:read_only]
      hrf.set(updated_at: DateTime.now, required_documents: hrf.required_documents)
      hrf.reload
    end

    hr.update_tax_revenues
    validate_fields(hr)
    hr.update_form_status
    hr.update_request_status
    hr.request_status = 'INCOMPLETE' if old_request_status == 'INCOMPLETE'
    hr.update_eligibility_query
    hr.updated_at = DateTime.now

    keys_to_update = %i[
      updated_at
      reference_tax_revenues
      request_status
      form_status
      form_front_validation_errors
      form_sne_validation_errors
      eligibility_query_base
      eligibility_query_department
      eligibility_query_seeked
      eligibility_query_bordering
      eligibility_query_seeked_bordering
      eligibility_query_other_from_department
    ]
    
    modified_fields = hr.changes&.keys.map(&:to_s) & keys_to_update.map(&:to_s)

    unless @options[:read_only]
      data = {}
      keys_to_update.each do |key|
        data[key] = hr[key]
      end
      hr.set(data)
      hr.reload
      hrf&.reload
    end

    apps = Hermes::Applicant.only(
      :id, 
      :year_minus_two,
      :amount_year_minus_two,
      :year_minus_one,
      :amount_year_minus_one,
    ).where(:id.in => all_applicant_ids)

    log_line = [
      hr.id,
      hr.als_id,
      hr.unique_numbers&.last,
      hr.updated_at,
      hr.active_application_id,
      hr.main_applicant_id,
      apps.map { |a| "#{a.id} (#{a.dynamic_year_minus_two}): #{a.dynamic_amount_minus_two}"}&.join(', '),
      apps.map { |a| "#{a.id} (#{a.dynamic_year_minus_one}): #{a.dynamic_amount_minus_one}"}&.join(', '),
      apps.map { |a| "#{a.id} (#{a.year_minus_two}): #{a.amount_year_minus_two}"}&.join(', '),
      apps.map { |a| "#{a.id} (#{a.year_minus_one}): #{a.amount_year_minus_one}"}&.join(', '),
      old_reference_tax_revenues,
      hr.reference_tax_revenues,
      old_request_status,
      hr.request_status,
      old_required_document_status&.join(', '),
      get_required_document_status(hrf)&.join(', '),
    ]

    add_log_line(log_line, return_code, modified_fields, code_details)
  rescue StandardError => e
    return_code = 'KO'
    code_details = "Error : #{e}"
    log_line = [record.id]
    modified_fields = nil

    add_log_line(log_line, return_code, modified_fields, code_details)
  end
end

# Run the script all options are there
ScriptRunner.new(
  read_only: false,
  skip: 0,
).run
